"""Created an example mc experiment (goes with estimate_py.yaml) to show the usage of the mc experiments."""

import logging
from typing import Dict
import numpy as np

from experiment_lab.experiments.monte_carlo import UniformSampler, BasePostProcessor

logger = logging.getLogger(__name__)


class RandomPointInCircle(UniformSampler):
    """Sampler class to sample from a bernoulli distribution with p=\\pi / 4."""

    def __init__(self) -> None:
        """Constructor for RandomPointInCircle sampler."""
        super().__init__(low=0, high=1, k=2)

    def sample(self, num_samples: int = 1) -> np.ndarray:
        """Samples random points from the box between the points (0, 0) and (1, 1) and then checks if they within the circle with radius 1 centered at the origin.

        Args:
            num_samples (int, optional): The number of samples. Defaults to 1.

        Returns:
            np.ndarray: The binary samples.
        """
        return (np.linalg.norm(super().sample(num_samples), axis=1) < 1).astype(int)


class MultiplyByFour(BasePostProcessor):
    """A post processing class to multiply the mean by four to get the pi estimate."""

    def __init__(self) -> None:
        """Constructor for MultiplyByFour post processor."""
        super().__init__()

    def process(self, result: Dict[str, float]) -> Dict[str, float]:
        """Post processing function that takes the results from the multi aggregator and calculates the pi estimate (multiply by 4).

        Args:
            result (Dict[str, float]): The aggregated result with the mean in it.

        Returns:
            Dict[str, float]: The result containing the pi estimate and everything else in it.
        """
        return {"pi_estimate": result["mean"] * 4, **result}



"Random wait experiment python file."
import time
import numpy as np
import wandb
import logging

from experiment_lab.core.base_config import BaseConfig
from experiment_lab.core.base_experiment import BaseExperiment
from experiment_lab.core.runner import run_experiment

logger = logging.getLogger(__name__)


class RandomWaits(BaseExperiment):
    """The random wait experiment class."""

    def __init__(self, cfg: BaseConfig) -> None:
        """The constructor for the random wait experiment

        Args:
            cfg (BaseConfig): The experiment configuration.
        """
        assert type(cfg) == BaseConfig
        super().__init__(cfg)

    def single_run(
        self, run_id: str = "", run_output_path: str = "", seed: int | None = None
    ) -> None:
        """Runs the trivial random wait experiment a single time with one seed.

        Args:
            run_id (str, optional): The run id. Defaults to "".
            seed (int | None, optional): The seed. Defaults to None.

        Returns:
            Any: Any resulting metrics
        """
        logger.info("Starting")
        rng = np.random.default_rng(seed=seed)
        num = rng.integers(10)
        if self.cfg.wandb:
            wandb.log({"num": num})
        time.sleep(num)
        logger.info("Done!")


if __name__ == "__main__":
    run_experiment(experiment_cls=RandomWaits)



"""Python file for the aggregators."""

import abc
from typing import Any, Callable, Dict
import hydra
import numpy as np


class BaseAggregator(abc.ABC):
    """The base aggregator class."""

    def __init__(self) -> None:
        """The constructor for a generic aggregator."""
        pass

    @abc.abstractmethod
    def aggregate(self, samples: np.ndarray) -> Any:
        """The aggregate function.

        Args:
            samples (np.ndarray): The samples to aggregate.

        Returns:
            Any: The aggregated result.
        """
        pass


class NpAggregator(BaseAggregator):
    """Aggregator to apply a numpy function to the axis=0 of the samples."""

    def __init__(self, np_func: Callable[..., np.ndarray] | str = np.mean) -> None:
        """The constructor of the NpAggregator that sets the np function to use.

        Args:
            np_func (Callable, optional): The np function to use. Defaults to np.mean.
        """
        super().__init__()
        if not isinstance(np_func, str):
            self.np_func: Callable[..., np.ndarray] = np_func
        else:
            self.np_func: Callable[..., np.ndarray] = hydra.utils.get_method(f"numpy.{np_func}")

    def aggregate(self, samples: np.ndarray) -> Any:
        """The aggregate function that applies the np function to the samples.

        Args:
            samples (np.ndarray): The samples.

        Returns:
            Any: The aggregated result.
        """
        return self.np_func(samples, axis=0)


class MultipleAggregators(BaseAggregator):
    """Aggregator to use multiple aggregators together."""

    def __init__(self, aggregators: Dict[str, BaseAggregator]) -> None:
        """Constructor for the aggregator to combine multiple aggregators.

        Args:
            aggregators (Dict[str, BaseAggregator]): The aggregators to combine.
        """
        super().__init__()
        self.aggregators = aggregators

    def aggregate(self, samples: np.ndarray) -> Dict[str, Any]:
        """Aggregates the samples using each aggregators and combines the results.

        Args:
            samples (np.ndarray): The samples to aggregate.

        Returns:
            Dict[str, Any]: The aggregated results.
        """
        return {k: v.aggregate(samples=samples) for k, v in self.aggregators.items()}



"""Python file containing the configs for the supervised learning experiments."""

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict
from hydra.core.config_store import ConfigStore

from experiment_lab.core.base_config import BaseConfig


class IntervalType(Enum):
    """The log interval type for wandb and logging"""

    batches = 0
    epochs = 1
    batches_and_epochs = 2


@dataclass
class LoggingSchedule:
    log_type: IntervalType = IntervalType.batches
    log_freq: int = 0
    wandb_log_type: IntervalType = IntervalType.batches
    wandb_log_freq: int = 0


@dataclass
class SupervisedConfig(BaseConfig):
    """The Supervised dataclass."""

    dataset: Dict[str, Any] = field(
        default_factory=lambda: {
            "_target_": "torch.utils.data.ConcatDataset",
            "datasets": [
                {
                    "_target_": "torchvision.datasets.MNIST",
                    "root": "./data/supervised/mnist",
                    "download": True,
                    "train": True,
                    "transform": {"_target_": "torchvision.transforms.ToTensor"},
                },
                {
                    "_target_": "torchvision.datasets.MNIST",
                    "root": "./data/supervised/mnist",
                    "download": True,
                    "train": False,
                    "transform": {"_target_": "torchvision.transforms.ToTensor"},
                },
            ],
        }
    )

    train_ratio: float = 0.8
    val_ratio: float = 0.1
    test_ratio: float = 0.1

    data_loader_kwargs: Dict[str, Any] | None = None

    model: Dict[str, Any] = field(
        default_factory=lambda: {
            "_target_": "torch.nn.Sequential",
            "_args_": [
                {
                    "_target_": "torch.nn.Flatten",
                },
                {
                    "_target_": "experiment_lab.common.networks.create_mlp_network",
                    "layer_sizes": [784, 500, 100, 10],
                },
            ],
        }
    )
    load_model_path: str | None = None

    optimizer: Dict[str, Any] = field(
        default_factory=lambda: {
            "_target_": "torch.optim.Adam",
        }
    )

    scheduler: Dict[str, Any] | None = None

    loss: Dict[str, Any] = field(
        default_factory=lambda: {"_target_": "torch.nn.NLLLoss"}
    )

    num_epochs: int = 10

    save_model_freq: int = 1  # Always by num epochs
    logging_schedule: LoggingSchedule = field(default_factory=LoggingSchedule)

    device: str = "mps"
    gpu_idx: int | None = None

    def __post_init__(self) -> None:
        """Does additional checks on the loaded config."""
        super().__post_init__()
        assert self.num_epochs >= 0, "The number of epochs cannot be less than 0"
        assert self.train_ratio > 0, "The train ratio must be greater than 0"
        assert self.val_ratio > 0, "The val ratio must be greater than 0"
        assert self.test_ratio > 0, "The test ratio must be greater than 0"
        assert (
            self.train_ratio + self.val_ratio + self.test_ratio == 1.0
        ), "All of the data must be split up into train, val, and test such that train_ratio + val_ratio + test_ratio == 1.0"


def register_configs():
    """Registers the supervised config."""
    cs = ConfigStore.instance()
    cs.store(name="supervised_config", node=SupervisedConfig)



"""Python file for sample filters."""

import abc
import numpy as np


class BaseSampleFilter(abc.ABC):
    """The base sample filter class."""

    def __init__(self) -> None:
        """Constructor for a generic sample filter."""
        super().__init__()

    @abc.abstractmethod
    def filter_samples(self, samples: np.ndarray) -> np.ndarray:
        """Filters the un-aggregated samples.

        Args:
            samples (np.ndarray): The un-aggregated un-filtered samples.

        Returns:
            np.ndarray: The un-aggregated filtered samples.
        """
        pass


class PassThroughSampleFilter(BaseSampleFilter):
    """The pass through sample filter class."""

    def __init__(self) -> None:
        """Constructor for pass through sample filter."""
        super().__init__()

    def filter_samples(self, samples: np.ndarray) -> np.ndarray:
        """Returns the full sample list.

        Args:
            samples (np.ndarray): The samples.

        Returns:
            np.ndarray: The samples.
        """
        return samples



"""The file to hold the base configuration object."""

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List
from hydra.core.config_store import ConfigStore


class NRunMethodEnum(Enum):
    """Method to run multi runs (series or parallel)"""

    series = 0
    parallel = 1


@dataclass
class AnalysisConfig:
    """Config for analysis of this experiment."""

    filters: List[Dict] = field(default_factory=lambda: [])
    wandb_keys: List[str] = field(default_factory=lambda: [])
    index: str | None = None
    all_keys_per_step: bool = False
    load_from_output_dir: bool = False
    output_dir_to_load_from: str | None = None


@dataclass
class BaseConfig:
    """Basic configuration for general experiment runs."""

    experiment_name: str | None = None

    seed: int | None = None

    n_runs: int = 1
    n_run_method: NRunMethodEnum = NRunMethodEnum.series

    wandb: Dict[str, Any] | None = None

    run_analysis: bool = False
    analysis: AnalysisConfig = field(default_factory=AnalysisConfig)

    def __post_init__(self) -> None:
        """Validation checks for base config"""
        assert self.n_runs >= 0, "Number of runs must be at least 0."


def register_configs() -> None:
    """Registers the configs created."""
    cs = ConfigStore.instance()
    cs.store(name="base_config", node=BaseConfig)



"""The file to hold the base experiment code."""

from dataclasses import asdict
from typing import Any, Sequence, Tuple

import abc
import os
import logging
import multiprocessing as mp
from hydra.core.hydra_config import HydraConfig
from omegaconf import OmegaConf
import wandb
import time
import glob
import pickle as pkl

from experiment_lab.common.utils import time_f, camel_to_snake_case
from experiment_lab.core.base_config import BaseConfig, NRunMethodEnum

logger = logging.getLogger(__name__)


class BaseExperiment(abc.ABC):
    """A generic base experiment class for all experiments to inherit from."""

    INCR_SEED_BY = 12

    def __init__(self, cfg: BaseConfig) -> None:
        """The constructor for the base experiment class."""
        self.cfg = cfg
        self.initialize_experiment()

    def initialize_experiment(self) -> None:
        """Initializes the experiment that is about to get run."""
        self.wandb_dict_config = asdict(self.cfg)
        if self.cfg.experiment_name:
            self.experiment_name = self.cfg.experiment_name
        else:
            self.experiment_name = (
                f"{camel_to_snake_case(self.__class__.__name__.lower())}"
            )
        self.timestamp = int(time.time())
        self.experiment_id = f"{self.experiment_name}_{self.timestamp}"
        self.output_directory = HydraConfig.get().runtime.output_dir
        self.additional_wandb_init_kwargs = {}

    @abc.abstractmethod
    def single_run(
        self, run_id: str, run_output_path: str, seed: int | None = None
    ) -> Any:
        """The entrypoint to the experiment.

        Args:
            run_id (str): The a unique string id for the run.
            run_output_path (str): The path to output or save anything for the run.
            seed (int | None, optional): The random seed to use for the experiment run. Defaults to None.

        Returns:
            Any: The results from this experiment run.
        """
        pass

    def _single_run_wrapper(
        self, intial_seed_and_run_num: Tuple[int | None, int] = (None, 0)
    ) -> Any:
        """A wrapper for the single run call to do run specific setup and post processing.

        Args:
            intial_seed_and_run_num (Tuple[int  |  None, int], optional): A tuple with both the seed and the run number. Defaults to (None, 0).

        Returns:
            Any: The results from the experiment run.
        """
        seed, run_num = intial_seed_and_run_num
        seed = None if seed is None else seed + run_num * BaseExperiment.INCR_SEED_BY
        logger.info(f"Starting individual run with seed {seed}")
        start_ns = time.time_ns()
        wandb_run = None
        run_id = f"{self.experiment_id}_{run_num}_{self.cfg.n_runs}"
        run_output_path = f"{self.output_directory}/{run_id}"
        os.makedirs(run_output_path, exist_ok=True)

        if self.cfg.wandb:
            wandb_run = wandb.init(
                id=run_id,
                config={
                    **self.wandb_dict_config,
                    "experiment_name": self.experiment_name,
                    "timestamp": self.timestamp,
                    "experiment_id": self.experiment_id,
                    "experiment_cls": self.__class__.__name__.lower(),
                },
                reinit=True,
                settings=wandb.Settings(start_method="thread"),
                **self.additional_wandb_init_kwargs,
                **self.cfg.wandb,
            )
        result = self.single_run(
            run_id=run_id, run_output_path=run_output_path, seed=seed
        )
        if result is not None:
            with open(f"{run_output_path}/result.pkl", "wb") as f:
                pkl.dump(result, f)
        end_ns = time.time_ns()
        logger.info(
            f"Finished run with seed {seed}. Time elapsed: {(end_ns - start_ns) / 1e9}s"
        )
        if wandb_run is not None:

            def save_folder(folder_path, base_path):
                for p in glob.glob(f"{folder_path}/**", recursive=True):
                    if os.path.isfile(p):
                        wandb_run.save(p, base_path=base_path, policy="end")

            save_folder(folder_path=run_output_path, base_path=run_output_path)
            save_folder(
                folder_path=f"{self.output_directory}/.hydra",
                base_path=self.output_directory,
            )
            wandb_run.finish()
        return result

    @time_f
    def run(self) -> Sequence[Any]:
        """Runs the experiment multiple times in series and aggregates the results.

        Args:
            n_runs (int, optional): The number of times to run the experiment. Defaults to 1.
            seed (int | None, optional): The initial seed of the first experiment run. The seeds of further experiment runs increment by 1 each run. Defaults to None.

        Returns:
            Sequence[Any]: The list of results from the runs.
        """
        results = None
        if self.cfg.n_runs <= 1 or self.cfg.n_run_method == NRunMethodEnum.series:
            results = [
                self._single_run_wrapper((self.cfg.seed, run_num))
                for run_num in range(self.cfg.n_runs)
            ]
        elif self.cfg.n_run_method == NRunMethodEnum.parallel:
            with mp.Pool() as pool:
                results = pool.map(
                    func=self._single_run_wrapper,
                    iterable=[
                        (self.cfg.seed, run_num) for run_num in range(self.cfg.n_runs)
                    ],
                )
        assert results is not None, "Unknown n_run_method!"
        return results



"""Runs the analysis for an rl experiment."""

from typing import Any, Dict

from pandas import DataFrame
from experiment_lab.core.base_analysis import BaseAnalysis
from experiment_lab.experiments.rl.config import RLConfig


class RLAnalysis(BaseAnalysis):
    """The class for analysis of rl experiments"""

    def __init__(self, cfg: RLConfig) -> None:
        """Initialization for rl analysis class

        Args:
            cfg (RLConfig): The rl config from the experiment.
        """
        self.cfg = cfg

    def analyze(self, df: DataFrame, configs: Dict[str, Dict[str, Any]]) -> Any:
        """The main analysis function for an rl experiment.

        Args:
            df (DataFrame): The run data from the experiments.
            configs (Dict[str, Dict[str, Any]]): The configs for each experiment.

        Returns:
            Any: The results of the analysis.
        """
        results = {
            "max_rewards": {
                experiment_id: {
                    "mean": df.loc[experiment_id]["rollout/ep_rew_mean"]
                    .groupby("run_id")
                    .max()
                    .mean(),
                    "std": df.loc[experiment_id]["rollout/ep_rew_mean"]
                    .groupby("run_id")
                    .max()
                    .std(),
                }
                for experiment_id in configs
            }
        }
        return results



"""A python module a bunch of network building blocks."""

from typing import Any, Dict
import torch
from torch import nn


class Convolution(nn.Module):
    """Convolution block."""

    def __init__(
        self,
        conv_kwargs: Dict[str, Any],
        max_pooling_kwargs: Dict[str, Any],
        nd: int = 2,
    ) -> None:
        """Constructor for convolution block taking the conv kwargs and pool kwargs.

        Args:
            conv_kwargs (Dict[str, Any]): The kwargs for the torch convolution layer.
            max_pooling_kwargs (Dict[str, Any]): The kwargs for the torch max pooling layer.
            nd (int, optional): The number of dimensions to convolve over. Defaults to 2.

        Raises:
            ValueError: When nd is not 1, 2, or 3.
        """
        super().__init__()
        self.nd = nd
        if self.nd == 1:
            self.conv = nn.Conv1d(**conv_kwargs)
        elif self.nd == 2:
            self.conv = nn.Conv2d(**conv_kwargs)
        elif self.nd == 3:
            self.conv = nn.Conv3d(**conv_kwargs)
        else:
            raise ValueError("nd must be 1, 2, or 3 for convolution block.")

        if self.nd == 1:
            self.pool = nn.MaxPool1d(**max_pooling_kwargs)
        elif self.nd == 2:
            self.pool = nn.MaxPool2d(**max_pooling_kwargs)
        elif self.nd == 3:
            self.pool = nn.MaxPool3d(**max_pooling_kwargs)
        else:
            raise ValueError("nd must be 1, 2, or 3 for convolution block.")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """The forward function on the convolution block.

        Args:
            x (torch.Tensor): The input.

        Returns:
            torch.Tensor: The output.
        """
        return self.pool(self.conv(x))



"""The rl experiment python file."""

import os
from typing import Any, List, Type, Dict
import copy
import gymnasium as gym
import hydra
from omegaconf import OmegaConf
import torch
from wandb.integration.sb3 import WandbCallback

from stable_baselines3.common.callbacks import CallbackList
from stable_baselines3.common.policies import BasePolicy
from stable_baselines3.common.callbacks import BaseCallback
from stable_baselines3.common.vec_env import VecVideoRecorder


from experiment_lab.core.base_experiment import BaseExperiment
from experiment_lab.experiments.rl.config import RLConfig
from experiment_lab.experiments.rl.environment import GeneralVecEnv
from experiment_lab.common.utils import default


class RLExperiment(BaseExperiment):
    """The rl experiment class."""

    def __init__(self, cfg: RLConfig) -> None:
        """Constructor for rl experiment.

        Args:
            cfg (RLConfig): The config to run the experiment with.
        """
        self.cfg = cfg
        self.initialize_experiment()

    def initialize_experiment(self) -> None:
        """Initializes the rl experiment"""
        super().initialize_experiment()

        self.wrapper_cls_lst: List[Type[gym.Wrapper]] = (
            []
            if self.cfg.wrapper_cls_lst is None
            else [
                hydra.utils.get_class(wrapper_cls)
                for wrapper_cls in self.cfg.wrapper_cls_lst
            ]
        )
        self.policy_cls: Type[BasePolicy] = hydra.utils.get_class(self.cfg.policy_cls)
        self.callback_cls_lst: List[Type[BaseCallback]] = (
            []
            if self.cfg.callback_cls_lst is None
            else [
                hydra.utils.get_class(callback_cls)
                for callback_cls in self.cfg.callback_cls_lst
            ]
        )

        self.wrapper_kwargs_lst: List[Dict[str, Any]] = default(
            self.cfg.wrapper_kwargs_lst, []
        )
        for _ in range(len(self.wrapper_kwargs_lst), len(self.wrapper_cls_lst)):
            self.wrapper_kwargs_lst.append({})

        self.callback_kwargs_lst: List[Dict[str, Any]] = default(
            self.cfg.callback_kwargs_lst, []
        )
        for _ in range(len(self.callback_kwargs_lst), len(self.callback_cls_lst)):
            self.callback_kwargs_lst.append({})

        self.model_kwargs: Dict[str, Any] = default(self.cfg.model_kwargs, {})

        self.transfer_steps: List[int] = default(self.cfg.transfer_steps, [])

        self.device: torch.device = torch.device(self.cfg.device)

        self.additional_wandb_init_kwargs["sync_tensorboard"] = True
        self.additional_wandb_init_kwargs["save_code"] = True
        if self.cfg.record_policy_videos:
            self.additional_wandb_init_kwargs["monitor_gym"] = True

    def single_run(
        self, run_id: str, run_output_path: str, seed: int | None = None
    ) -> Any:
        """A single run of the rl experiment.

        Args:
            run_id (str): The run id.
            run_output_path (str): The run output path.
            seed (int | None, optional): The initial seed.. Defaults to None.

        Returns:
            Any: Any resulting metrics.
        """

        # Setup the device to train on
        if self.cfg.gpu_idx is not None:
            os.environ["CUDA_VISIBLE_DEVICES"] = str(self.cfg.gpu_idx)

        # Create the environment
        env = GeneralVecEnv(
            env_config=self.cfg.env_config,
            transfer_steps=self.cfg.transfer_steps,
            wrappers=self.wrapper_cls_lst,
            wrapper_kwargs_lst=self.cfg.wrapper_kwargs_lst,
            n_envs=self.cfg.n_envs,
            seed=seed,
            monitor_dir=self.cfg.monitor_dir,
            monitor_kwargs=self.cfg.monitor_kwargs,
            start_method=self.cfg.start_method,
            render_mode=(
                "rgb_array" if self.cfg.record_policy_videos else self.cfg.render_mode
            ),
        )
        if self.cfg.record_policy_videos:
            env = VecVideoRecorder(
                env,
                f"./wandb/videos/{run_id}",
                record_video_trigger=lambda x: x
                % (self.cfg.video_freq // self.cfg.n_envs)
                == 0,
                video_length=self.cfg.video_length,
                name_prefix=self.cfg.video_name_prefix,
            )

        env.seed(seed)
        env.reset()

        # Setup the model
        model = hydra.utils.instantiate(
            {
                "_target_": self.cfg.model_cls,
                "policy": self.policy_cls,
                "policy_kwargs": self.cfg.policy_kwargs,
                **self.model_kwargs,
            },
            env=env,
            device=self.device,
            tensorboard_log=f"./logs/" if self.cfg.log else None,
            verbose=self.cfg.verbose,
            _convert_="partial",
        )
        model.set_random_seed(seed)
        model_save_path = f"{run_output_path}/models/"

        # Instantiate the callbacks
        callback_instances = []
        if (
            self.cfg.log
            and self.cfg.wandb
            and self.cfg.wandb_callback_kwargs is not None
        ):
            callback_instances.append(
                WandbCallback(
                    gradient_save_freq=self.cfg.wandb_callback_kwargs.gradient_save_freq,
                    model_save_freq=self.cfg.wandb_callback_kwargs.model_save_freq,
                    model_save_path=model_save_path,
                    verbose=self.cfg.verbose,
                )
            )
        for callback_cls, callback_kwargs in zip(
            self.callback_cls_lst, self.callback_kwargs_lst
        ):
            callback_instances.append(
                hydra.utils.instantiate(
                    {
                        "_target_": callback_cls,
                        **callback_kwargs,
                    }
                )
            )

        # Run the algorithm
        model.learn(
            total_timesteps=self.cfg.total_time_steps,
            log_interval=self.cfg.log_interval,
            tb_log_name=run_id,
            callback=(
                CallbackList(callback_instances)
                if len(callback_instances) > 0
                else None
            ),
        )

        env.close()

        if self.cfg.save_model:
            os.makedirs(model_save_path, exist_ok=True)
            model.save(f"{model_save_path}/final.zip")



"""Main entrypoint to run supervised learning experiments."""

from experiment_lab.core.runner import run_experiment, root_config_folder
from experiment_lab.experiments.supervised.experiment import SupervisedExperiment
from experiment_lab.experiments.supervised.config import (
    SupervisedConfig,
    register_configs,
)

from functools import partial

run_supervised_experiment = partial(
    run_experiment,
    experiment_cls=SupervisedExperiment,
    config_cls=SupervisedConfig,
    register_configs=register_configs,
    config_path=f"{root_config_folder}/supervised",
)


if __name__ == "__main__":
    run_supervised_experiment()



"""Python file for the samplers."""

import abc
import numpy as np


class BaseSampler(abc.ABC):
    """The base sampler class."""

    def __init__(self) -> None:
        """Constructor for a generic sampler."""
        super().__init__()
        self.rng = np.random.default_rng()

    def set_seed(self, seed: int | None):
        """Sets the seed for the internal random number generator.

        Args:
            seed (int | None): The seed to set.
        """
        self.seed = seed
        self.rng = np.random.default_rng(seed=seed)

    @abc.abstractmethod
    def sample(self, num_samples: int = 1) -> np.ndarray:
        """Samples a single point from the distribution represented by this sampler.

        Args:
            num_samples (int): The number of samples to generate.

        Returns:
            np.ndarray: The sample represented as a numpy array with shape (num_samples, ...)
        """
        pass


class UniformSampler(BaseSampler):
    """Samples from a uniform distribution"""

    def __init__(self, low: float = 1.0, high: float | None = None, k: int = 1) -> None:
        """Constructor for the uniform sampler.

        Args:
            low (float, optional): The minimum of the distribution to sample from (or maximum if high is left as None, in which case low is 0). Defualts to 1.0.
            high (float | None, optional): The maximum of the distribution to sample from. Defaults to None.
            k (int, optional): The number of random numbers per sample. Defaults to 1.
        """

        super().__init__()
        if high is None:
            self.low = 0
            self.high = low
        else:
            self.low = low
            self.high = high
        self.k = k

    def sample(self, num_samples: int = 1) -> np.ndarray:
        return self.rng.uniform(
            low=self.low, high=self.high, size=(num_samples, self.k)
        )



"""The monte carlo experiment file."""

import logging
from typing import Any
import hydra
import numpy as np
import wandb
from experiment_lab.core.base_experiment import BaseExperiment
from experiment_lab.experiments.monte_carlo.components import (
    BaseSampler,
    BaseAggregator,
    BaseSampleFilter,
    BasePostProcessor,
)
from experiment_lab.experiments.monte_carlo.config import MCConfig


logger = logging.getLogger(__name__)


class MCExperiment(BaseExperiment):
    """The monte carlo experiment class."""

    def __init__(self, cfg: MCConfig) -> None:
        """The constructor for the monte carlo experiment."""
        self.cfg = cfg
        self.initialize_experiment()

    def initialize_experiment(self) -> None:
        """Initializes the monte carlo experiment config."""
        super().initialize_experiment()
        self.batch_size = (
            self.cfg.n_samples if self.cfg.batch_size is None else self.cfg.batch_size
        )

    def single_run(
        self, run_id: str, run_output_path: str, seed: int | None = None
    ) -> Any:
        """A single run of the monte carlo experiment.

        Args:
            run_id (str): The run id.
            run_output_path (str): The run output path.
            seed (int | None, optional): The initial random seed. Defaults to None.

        Returns:
            Any: The post processed results.
        """

        # Instantiate the required experiment components
        sampler: BaseSampler = hydra.utils.instantiate(self.cfg.sampler)
        aggregator: BaseAggregator = hydra.utils.instantiate(self.cfg.aggregator)
        sample_filter: BaseSampleFilter = hydra.utils.instantiate(
            self.cfg.sample_filter
        )
        post_processor: BasePostProcessor = hydra.utils.instantiate(
            self.cfg.post_processor
        )

        # Set the seed
        sampler.set_seed(seed=seed)

        # Sample, Filter, Aggregate, and Post Process, all while logging
        all_samples = None
        results = None
        for i in range(0, self.cfg.n_samples, self.batch_size):
            new_samples = sampler.sample(min(self.batch_size, self.cfg.n_samples - i))
            filtered_samples = sample_filter.filter_samples(new_samples)
            if all_samples is None:
                all_samples = filtered_samples
            else:
                all_samples = np.concatenate([all_samples, filtered_samples], axis=0)
            if self.cfg.aggregate_every_batch:
                if results is None:
                    results = []
                results.append(
                    post_processor.process(aggregator.aggregate(all_samples))
                )
                if self.cfg.wandb:
                    wandb.log(results[-1])
                if self.cfg.log_results:
                    logger.info(f"The current result is: {results[-1]}")
        if not self.cfg.aggregate_every_batch and all_samples is not None:
            results = post_processor.process(aggregator.aggregate(all_samples))
            if self.cfg.wandb:
                wandb.log(results)
            if self.cfg.log_results:
                logger.info(f"The current result is: {results}")
        return results



"""A python module for utilities functions."""

import collections.abc
from typing import Any, Callable, Dict
import time
import logging
import re

logger = logging.getLogger(__name__)


def default(val: Any, d: Any, null_val: Any = None) -> Any:
    """Defaults a value to something when it matches a null_val.

    Args:
        val (Any): The original value.
        d (Any): The value to set if val == null_val.
        null_val (Any, optional): The null value. Defaults to None.

    Returns:
        Any: The value defaulting to the chosen default.
    """
    return val if val != null_val else d


def deep_update(d: Dict[Any, Any], u: collections.abc.Mapping) -> Dict[Any, Any]:
    """Deep update for dictionaries.

    Args:
        d (Dict[Any, Any]): The initial dictionary.
        u (collections.abc.Mapping): The mapping to update the initial dictionary with.

    Returns:
        Dict[Any, Any]: The updated dictionary.
    """
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def time_f(f: Callable) -> Callable:
    def g(*args, **kwargs):
        logger.info(f"Start {f.__name__}")
        start_ns = time.time_ns()
        result = f(*args, **kwargs)
        end_ns = time.time_ns()
        logger.info(f"End {f.__name__}. Time Elapsed: {(end_ns - start_ns) / 1e9}s")
        return result

    return g

def camel_to_snake_case(s: str) -> str:
    return re.sub(r'(?<!^)(?=[A-Z])', '_', s).lower()



"""The environment wrapper for gym to run the rl experiment."""

from typing import Any, List, SupportsFloat, Tuple, Dict, Type

import os

import gymnasium as gym
from gymnasium.core import RenderFrame
from gymnasium.envs.registration import EnvSpec
import hydra
import numpy as np
import multiprocessing as mp

from stable_baselines3.common.monitor import Monitor
from stable_baselines3.common.vec_env import SubprocVecEnv
from stable_baselines3.common.vec_env.base_vec_env import VecEnvStepReturn


class ListEnv(gym.Env):
    """An environment for maintaining multiple transfers."""

    def __init__(self, env_lst: List[gym.Env]) -> None:
        """The constructor for the list env class.

        Args:
            env_lst (List[gym.Env]): The list of environments to use.
        """
        self.env_lst = env_lst
        self.env_idx = 0

    def incr_env_idx(self) -> bool:
        """Increments the environment to the next

        Returns:
            bool: Whether or not an environment switch occured.
        """
        if self.env_idx >= len(self.env_lst) - 1:
            return False
        self.cur_env.close()
        self.env_idx += 1
        self.cur_env.reset()
        return True

    def step(
        self, action: Any
    ) -> Tuple[Any, SupportsFloat, bool, bool, Dict[str, Any]]:
        """Steps through the current environment.

        Args:
            action (Any): The action to take

        Returns:
            Tuple[Any, SupportsFloat, bool, bool, Dict[str, Any]]: The step output.
        """
        return self.cur_env.step(action=action)

    def reset(
        self, *, seed: int | None = None, options: Dict[str, Any] | None = None
    ) -> Tuple[Any, Dict[str, Any]]:
        """Resets the current environment.

        Args:
            seed (int | None, optional): The seed to use. Defaults to None.
            options (Dict[str, Any] | None, optional): The options to use. Defaults to None.

        Returns:
            Tuple[Any, Dict[str, Any]]: The observation and info in the reset environment.
        """
        return self.cur_env.reset(seed=seed, options=options)

    def render(self) -> RenderFrame | List[RenderFrame] | None:
        """Renders the current environment.

        Returns:
            RenderFrame | List[RenderFrame] | None: The render return.
        """
        return self.cur_env.render()

    def close(self):
        """Closes all the environments."""
        for env in self.env_lst:
            env.close()

    @property
    def cur_env(self) -> gym.Env:
        """Gets the current environment.

        Returns:
            gym.Env: The current environment.
        """
        return self.env_lst[self.env_idx]

    @property
    def unwrapped(self) -> gym.Env:
        """Returns the base non-wrapped environment.

        Returns:
            Env: The base non-wrapped :class:`gymnasium.Env` instance
        """
        return self.cur_env.unwrapped

    @property
    def action_space(self) -> gym.Space:
        """Returns the action space of the current environment.

        Returns:
            gym.Space: The action space.
        """
        return self.cur_env.action_space

    @property
    def observation_space(self) -> gym.Space:
        """Returns the observation space of the current environment.

        Returns:
            gym.Space: The observation space.
        """
        return self.cur_env.observation_space

    @property
    def reward_range(self) -> Tuple[float, float]:
        """The reward range of the environment.

        Returns:
            Tuple[float, float]: The reward range.
        """
        return self.cur_env.reward_range

    @property
    def spec(self) -> EnvSpec | None:
        return self.cur_env.spec

    @property
    def np_random(self) -> np.random.Generator:
        """The np_random generator of the current env.

        Returns:
            np.random.Generator: The generator.
        """
        return self.cur_env.np_random

    @property
    def render_mode(self) -> str | None:
        """The render mode of the current environment.

        Returns:
            str | None: The render mode.
        """
        return self.cur_env.render_mode


class GeneralVecEnv(SubprocVecEnv):
    """A full vec env wrapper around gym envs for easy construction and configuration."""

    def __init__(
        self,
        env_config: Dict[str, Any],
        transfer_steps: List[int] | None = None,
        wrappers: List[Type[gym.Wrapper]] | None = [],
        wrapper_kwargs_lst: List[Dict[str, Any]] | None = [],
        n_envs: int = 1,
        seed: int | None = None,
        monitor_dir: str | None = None,
        monitor_kwargs: Dict[str, Any] | None = None,
        start_method: str | None = None,
        render_mode: str | None = None,
    ) -> None:
        """Constructor for general vec env wrapper for the rl experiment module.

        Args:
            env_config (Dict[str, Any]): The environment config to use.
            transfer_steps (List[int] | None, optional): The number of steps to run before each transfer specified in the environment config. Defaults to None.
            wrappers (List[Type[gym.Wrapper]] | None, optional): The wrappers to apply to the env. Defaults to [].
            wrapper_kwargs_lst (List[Dict[str, Any]] | None, optional): The kwargs for the wrappers. Defaults to [].
            n_envs (int, optional): The number of envs to run in parallel. Defaults to 1.
            seed (int | None, optional): The seed to use. Defaults to None.
            monitor_dir (str | None, optional): The directory to save monitor logs to. Defaults to None.
            monitor_kwargs (Dict[str, Any] | None, optional): The kwargs for the monitor. Defaults to None.
            start_method (str | None, optional): The start method used for multiprocessing envs. Defaults to None.
            render_mode (str | None, optional): The render mode of the environment. Defaults to None.
        """

        self.n_tasks = max(
            (len(env_config[k]) if type(env_config[k]) == list else 1)
            for k in env_config
        )
        env_configs = []
        for i in range(self.n_tasks):
            env_configs.append(
                {
                    k: ((v[i] if i < len(v) else v[-1]) if type(v) == list else v)
                    for k, v in env_config.items()
                }
            )

        for cfg in env_configs:
            if isinstance(cfg["env_id"], str) and cfg["env_id"] not in gym.registry:
                cfg["env_id"] = hydra.utils.get_class(cfg["env_id"])

        env_spec_mapping = {
            cfg["env_id"]: gym.registry[cfg["env_id"]]
            for cfg in env_configs
            if type(cfg["env_id"]) == str
        }

        self.n_envs = n_envs
        self.transfer_steps = [] if transfer_steps is None else transfer_steps

        self.total_time_steps = 0
        self.last_incr = 0
        self.cur_env_idx = 0

        monitor_kwargs = {} if monitor_kwargs is None else monitor_kwargs

        wrappers = [] if wrappers is None else wrappers
        wrapper_kwargs_lst = [] if wrapper_kwargs_lst is None else wrapper_kwargs_lst
        for _ in range(len(wrapper_kwargs_lst), len(wrappers)):
            wrapper_kwargs_lst.append({})

        def make_env_fn(rank):
            def _make_env(config):
                env_id = config["env_id"]
                env_kwargs = hydra.utils.instantiate(
                    {k: v for k, v in config.items() if k != "env_id"}
                )

                # Initialize the environment
                if isinstance(env_id, str):
                    env: gym.Env = gym.make(
                        env_spec_mapping[env_id], render_mode=render_mode, **env_kwargs
                    )
                else:
                    env: gym.Env = env_id(**env_kwargs, render_mode=render_mode)

                # Optionally use the random seed provided
                if seed is not None:
                    env.np_random = np.random.default_rng(seed=seed + rank)
                    env.action_space.seed(seed + rank)

                # Wrap the env in a Monitor wrapper
                # to have additional training information
                monitor_path = (
                    os.path.join(monitor_dir, str(rank))
                    if monitor_dir is not None
                    else None
                )
                # Create the monitor folder if needed
                if monitor_path is not None:
                    os.makedirs(monitor_path, exist_ok=True)
                env = Monitor(
                    env,
                    filename=monitor_path,
                    **hydra.utils.instantiate(monitor_kwargs),
                )

                # Wrap the environment with the provided wrappers
                for wrapper_cls, wrapper_kwargs in zip(wrappers, wrapper_kwargs_lst):
                    env = wrapper_cls(env, **hydra.utils.instantiate(wrapper_kwargs))

                return env

            def _init() -> gym.Env[Any, Any]:
                # Returns a list env with each env constructed from the config in env_configs
                return ListEnv([_make_env(config) for config in env_configs])

            return _init

        env_fns = [make_env_fn(rank=i) for i in range(n_envs)]

        if start_method is None:
            start_method = "spawn" if "spawn" in mp.get_all_start_methods() else None

        self.eval_env: ListEnv = make_env_fn(rank=n_envs)() # type: ignore

        super().__init__(env_fns=env_fns, start_method=start_method)

    def step_wait(self) -> VecEnvStepReturn:
        """Waits for the results of the step call and incremenets the environment idx if needed.

        Returns:
            VecEnvStepReturn: The step return.
        """
        observations, rewards, dones, infos = super().step_wait()
        # Increment total time steps
        self.total_time_steps += self.n_envs
        if (
            self.cur_env_idx < len(self.transfer_steps)
            and self.total_time_steps - self.last_incr
            > self.transfer_steps[self.cur_env_idx]
        ):
            self.last_incr = self.total_time_steps
            # Trigger the novelty if enough steps have passed
            transfer_injected: List[bool] = self.env_method("incr_env_idx")
            self.eval_env.incr_env_idx()
            if np.any(transfer_injected):
                self.cur_env_idx += 1
                dones[:] = True

        return observations, rewards, dones, infos
